/*! \file
 *  \brief Тестирование прямого двумерного БПФ для типа float
 */

#include "test.h"

int test_fft_float_2d(const int fft_min_size, const int fft_max_size)
{
    int error_count = 0;
    int freq1 = 1;
    int freq2 = 2;
    float *dst = memalign(fft_max_size * 2 * sizeof(float), fft_max_size * 2 * sizeof(float));
    float *src = memalign(fft_max_size * 2 * sizeof(float), fft_max_size * 2 * sizeof(float));
    float *fetalon = memalign(fft_max_size * 2 * sizeof(float), fft_max_size * 2 * sizeof(float));

    char *test_info = malloc(sizeof(char) * 1000000);
    int print_bytes;

    FILE *res_2d;
    if (MEM_TYPE == EXT_EXT)
        res_2d = fopen("test_FLOAT_FFT_2D_EXT_EXT.csv", "w");
    else
        res_2d = fopen("test_FLOAT_FFT_2D_EXT_INT.csv", "w");

    fprintf(res_2d, "%s,%s,%s,%s\n", "SIZE", "TIC_CNTR", "INSTR_CNTR", "STATUS");

    print_bytes = sprintf(test_info, "2d float FFT ");
    if (MEM_TYPE == EXT_EXT)
        print_bytes += sprintf(test_info + print_bytes, "EXT_EXT\n\n");
    else
        print_bytes += sprintf(test_info + print_bytes, "EXT_INT\n\n");

    int length, height;

    for (int size = fft_min_size; size <= fft_max_size; size <<= 1) {
        length = 1 << (int)log2(size) / 2;
        height = size / length;
        print_bytes += sprintf(test_info + print_bytes, "\nheight = %d, length = %d\n",
                                height, length);
        fprintf(res_2d, "%dx%d,", height, length);

        fft_t fft_obj;
        fft_obj.src = src;
        fft_obj.dst = dst;
        fft_obj.int_mem = (enum CALC_MEM)MEM_TYPE;
        fft_obj.dir = 1;
        fft_obj.dim = (enum Dimension)FFT_DIM;

        create_spectrum_2d(src, height, length, freq1, freq2, FLOAT);

        if (fft_plan_2d(&fft_obj, height, length, FLOAT) == -1) {
            error_count = 1;
            return error_count;
        };

        fft_execute(&fft_obj);

        memset(fetalon, 0.0, 2 * fft_max_size * sizeof(float));
        if (fft_obj.int_mem && fft_obj.big_size)
            fetalon[2 * (freq2 * height + freq1) + 1] = (float)(size);
        else
            fetalon[2 * (freq1 * length + freq2) + 1] = (float)(size);

        print_bytes += sprintf(test_info + print_bytes, "\nticks_fft = %d, instr_fft = %d",
                                fft_obj.ticks, fft_obj.instrs);
        fprintf(res_2d, "%d,%d,", fft_obj.ticks, fft_obj.instrs);
        float norm = maximum_norm(src, fetalon, fft_obj.size);
        print_bytes += sprintf(test_info + print_bytes, ", norm = %f ", norm);
        if (norm < 1e-1) {
            print_bytes += sprintf(test_info + print_bytes, " : OK\n\n");
            fprintf(res_2d, "%s\n", "passed");
        } else {
            ++error_count;
            print_bytes += sprintf(test_info + print_bytes, " : WRONG\n\n");
            fprintf(res_2d, "%s\n", "failed");
        }

        fft_free_sources(&fft_obj);
    }

    printf("%s", test_info);

    free(test_info);
    free(src);
    free(dst);
    free(fetalon);
    fclose(res_2d);
    return error_count;
}
